@@ -2,10 +2,10 @@
 
 #' Generate an access token for the Customer Journey Analytics API
 #'
-#' **Note:** `cja_auth()` is the primary function used for authorization. `auth_oauth()`
+#' **Note:** `cja_auth()` is the primary function used for authorization. `auth_s2s()`
 #' and `auth_jwt()` should typically not be called directly.
 #'
-#' @param type Either 'jwt' or 'oauth'. This can be set explicitly, but a best practice is
+#' @param type Either 'jwt' or 's2s' (default). This can be set explicitly, but a best practice is
 #' to run `cja_auth_with()` to set the authorization type as an environment variable before
 #' running `cja_auth()`
 #' @param ... Additional arguments passed to auth functions.
@@ -13,11 +13,13 @@
 #' @param client_secret The client secret, defined by a global variable or manually defined
 #' @param file A JSON file containing service account credentials required for JWT
 #' authentication. This file can be downloaded directly from the Adobe Console,
-#' and should minimally have the fields `API_KEY`, `CLIENT_SECRET`, `ORG_ID`,
+#' and should minimally have the fields `API_KEY` or `CLIENT_ID`, `CLIENT_SECRET`, `ORG_ID`,
 #' and `TECHNICAL_ACCOUNT_ID`.
 #' @param private_key Filename of the private key for JWT authentication.
 #' @param jwt_token _(Optional)_ A custom, encoded, signed JWT claim. If used,
 #'   `client_id` and `client_secret` are still required.
+#' @param s2s_token _(Optional)_ A custom, encoded, signed JWT claim. If used,
+#'   `client_id` and `client_secret` are still required.
 #' @param use_oob if `FALSE`, use a local webserver for the OAuth dance.
 #'   Otherwise, provide a URL to the user and prompt for a validation code.
 #'   Defaults to the value of the `httr_oob_default` default, or TRUE if
@@ -29,16 +31,19 @@
 #' @family auth
 #' @aliases cja_auth auth_jwt auth_oauth
 #' @export
-cja_auth <- function(type = 'jwt', ...) {
+cja_auth <- function(type = 's2s', ...) {
 
   if (is.null(type)) {
     stop("Authentication type missing, please set an auth type with `cja_auth_with`")
   }
-  type <- match.arg(type, c("jwt", "oauth"))
+  type <- match.arg(type, c("jwt", "oauth", 's2s'))
+
+  cja_auth_with(type)
 
   switch(type,
          jwt = auth_jwt(...),
-         oauth = auth_oauth(...)
+         oauth = auth_oauth(...),
+         s2s = auth_s2s(...)
   )
 }
 
@@ -53,7 +58,7 @@ cja_auth <- function(type = 'jwt', ...) {
 #' `cja_auth_with` sets the type of authorization for the session. This is used
 #' as the default by `cja_auth()` when no specific option is given.
 #'
-#' @param type The authorization type: 'oauth' or 'jwt'
+#' @param type The authorization type: 's2s'
 #' @param path The location for the cached authorization token. It should be a
 #' directory, rather than a filename. If this option is not set, the current
 #' working directory is used instead. If the location does not exist, it will
@@ -68,11 +73,11 @@ cja_auth <- function(type = 'jwt', ...) {
 #' @rdname cja_auth_with
 #' @aliases cja_auth_with cja_auth_path cja_auth_name
 #' @export
-cja_auth_with <- function(type) {
+cja_auth_with <- function(type = 's2s') {
   if (missing(type)) return(getOption("cjar.auth_type"))
 
   if (!is.null(type)) {
-    type <- match.arg(type, c("oauth", "jwt"))
+    type <- match.arg(type, c("oauth", "jwt", "s2s"))
   }
 
   options(cjar.auth_type = type)
@@ -154,6 +159,10 @@ retrieve_cja_token <- function(...) {
       # thing to do for JWT
       .cjar$token$refresh()
     }
+  } else if(type == 's2s') {
+    if (!token$validate()) {
+      .cjar$token$refresh()
+    }
   }
 
   return(.cjar$token)
@@ -182,7 +191,7 @@ token_path <- function(...) {
 #'
 #' @param token An `httr` `reponse` object or `oauth2.0_token` object
 #'
-#' @return Either 'oauth' or 'jwt'
+#' @return Either 's2s' or 'oauth'
 #'
 #' @noRd
 token_type <- function(token) {
@@ -190,6 +199,8 @@ token_type <- function(token) {
     "oauth"
   } else if (inherits(token, "AdobeJwtToken")) {
     "jwt"
+  } else if (inherits(token, 'AdobeS2SToken')) {
+    's2s'
   } else if (is.null(token)) {
     NULL
   } else {
@@ -218,6 +229,7 @@ get_token_config <- function(client_id,
   switch(type,
          oauth = httr::config(token = token),
          jwt = httr::add_headers(Authorization = paste("Bearer", content(token$token)$access_token)),
+         s2s = httr::add_headers(Authorization = paste("Bearer", token$token$access)),
          stop("Unknown token type")
   )
 }
@@ -234,33 +246,51 @@ get_token_config <- function(client_id,
 #'
 #' @return List of length two with elements `client_id` and `client_secret`
 #' @noRd
-get_env_vars <- function() {
-  client_id <- .cjar$client_id
-  client_secret <- .cjar$client_secret
-  org_id <- .cjar$org_id
+get_env_vars <- function(type = cja_auth_with()) {
+  if(type == 'oauth' || type == 'jwt') {
+    client_id <- .cjar$client_id
+    client_secret <- .cjar$client_secret
+    org_id <- .cjar$org_id
 
-  if (is.null(client_id) | is.null(client_secret) | is.null(org_id)) {
-    client_id <- Sys.getenv("CJA_CLIENT_ID")
-    client_secret <- Sys.getenv("CJA_CLIENT_SECRET")
-    org_id <- Sys.getenv("CJA_ORG_ID")
+    if (is.null(client_id) | is.null(client_secret) | is.null(org_id)) {
+      client_id <- Sys.getenv("CJA_CLIENT_ID")
+      client_secret <- Sys.getenv("CJA_CLIENT_SECRET")
+      org_id <- Sys.getenv("CJA_ORG_ID")
+    }
+
+    if (client_id == "" | client_secret == "" | org_id == "") {
+      # env_vars <- c(client_id = client_id,
+      #               client_secret = client_secret,
+      #               org_id = org_id)
+      #
+      # missing_envs <- names(env_vars[env_vars == ""])
+
+      stop("Variable 'CJA_AUTH_FILE' not found but required for JWT authentication.\nSee `?cja_auth`",
+           call. = FALSE)
+    }
+
+    list(
+      client_id = client_id,
+      client_secret = client_secret,
+      org_id = org_id
+    )
+  } else if(type == 's2s') {
+    client_id <- .cjar$client_id
+    client_secret <- .cjar$client_secret
+    org_id <- .cjar$org_id
+
+    if (is.null(client_id) | is.null(client_secret) | is.null(org_id)) {
+      secrets <- jsonlite::fromJSON(Sys.getenv("CJA_AUTH_FILE"))
+      client_id <- secrets$CLIENT_ID
+      client_secret <-  secrets$CLIENT_SECRETS
+      org_id <- secrets$ORG_ID
+    }
+    list(
+      client_id = client_id,
+      client_secret = client_secret,
+      org_id = org_id
+    )
   }
-
-  if (client_id == "" | client_secret == "" | org_id == "") {
-    # env_vars <- c(client_id = client_id,
-    #               client_secret = client_secret,
-    #               org_id = org_id)
-    #
-    # missing_envs <- names(env_vars[env_vars == ""])
-
-    stop("Variable 'CJA_AUTH_FILE' not found but required for JWT authentication.\nSee `?cja_auth`",
-         call. = FALSE)
-  }
-
-  list(
-    client_id = client_id,
-    client_secret = client_secret,
-    org_id = org_id
-  )
 }
 
 
@@ -284,6 +314,8 @@ auth_jwt <- function(file = Sys.getenv("CJA_AUTH_FILE"),
 
   secrets <- jsonlite::fromJSON(file)
 
+  cja_auth_with('jwt')
+
   resp <- auth_jwt_gen(secrets = secrets, private_key = private_key, jwt_token = jwt_token)
 
 
@@ -297,7 +329,6 @@ auth_jwt <- function(file = Sys.getenv("CJA_AUTH_FILE"),
   .cjar$org_id <- secrets$ORG_ID
 }
 
-
 #' Generate the authorization response object
 #'
 #' @param secrets List of secret values, see `auth_jwt`
@@ -340,7 +371,6 @@ auth_jwt_gen <- function(secrets,
   token
 }
 
-
 #' Get an encoded, signed JWT token
 #'
 #' Gets a JWT token
@@ -423,6 +453,131 @@ AdobeJwtToken <- R6::R6Class("AdobeJwtToken", list(
   }
 ))
 
+# S2S ------------------------------------------------------------------
+
+#' @family auth
+#' @describeIn cja_auth Authenticate with S2S token
+#' @importFrom openssl read_key
+#' @export
+auth_s2s <- function(file = Sys.getenv("CJA_AUTH_FILE"),
+                     s2s_token = NULL,
+                     ...) {
+  if (file == "") {
+    stop("Variable 'CJA_AUTH_FILE' not found but required for default S2S authentication.\nSee `?cja_auth`")
+  }
+
+  secrets <- jsonlite::fromJSON(file)
+
+  resp <- auth_s2s_gen(secrets = secrets, s2s_token = s2s_token)
+
+  cja_auth_with('s2s')
+
+  # If successful
+  message("Successfully authenticated with S2S: access token valid until ",
+          as.POSIXct(resp$expires_at, origin = "1970-01-01"))
+
+  .cjar$token <- AdobeS2SToken$new(resp, secrets)
+  .cjar$client_id <- secrets$CLIENT_ID
+  .cjar$client_secret <- secrets$CLIENT_SECRETS
+  .cjar$org_id <- secrets$ORG_ID
+}
+
+#' Generate the authorization response object for S2S
+#'
+#' @param secrets List of secret values, see `auth_s2s`
+#' @param s2s_token Optional, a S2S token (e.g., a cached token)
+#'
+#' @noRd
+auth_s2s_gen <- function(secrets,
+                         s2s_token = NULL) {
+
+  stopifnot(is.character(secrets$CLIENT_ID))
+  stopifnot(is.character(secrets$CLIENT_SECRETS))
+  stopifnot(is.character(secrets$SCOPES))
+
+  if (any(c(secrets$CLIENT_ID, secrets$CLIENT_SECRETS) == "")) {
+    stop("Client ID or Client Secret not found.")
+  }
+
+  s2s_token <- get_s2s_token(s2s_token = s2s_token,
+                             client_id = secrets$CLIENT_ID,
+                             scopes = secrets$SCOPES,
+                             client_secrets = secrets$CLIENT_SECRETS
+                            )
+
+  token <- s2s_token
+  token
+}
+
+
+
+#' Get an encoded, signed S2S token
+#'
+#' Gets a S2S token
+#'
+#' @param S2S_token Optional, a S2S token
+#' @param client_id Client ID
+#' @param secrets Secrets, can be more than one, a vector if more than one or a string if only 1
+#' @param scopes Scopes as listed in the project
+#'
+#' @return A S2S token generated by httr2 oauth_flow_client_credentials
+#' @noRd
+get_s2s_token <- function(s2s_token = NULL,
+                          client_id,
+                          client_secrets,
+                          scopes = c("openid","AdobeID","additional_info.projectedProductContext","read_pc.acp","read_pc","read_pc.dma_tartan","additional_info","read_organizations","session")
+                          ) {
+  if (is.null(s2s_token)) {
+    s2s_client <- httr2::oauth_client(client_id,
+                                      'https://ims-na1.adobelogin.com/ims/token/v3',
+                                      secret = paste0(client_secrets, collapse = ','),
+                                      auth = 'body')
+
+    s2s_token <- httr2::oauth_flow_client_credentials(client = s2s_client,
+                                                      scope = paste0(scopes, collapse = ','),
+                                                      token_params = c(grant_type = 'client_credentials',
+                                                                       client_id = client_id))
+
+  }
+
+  s2s_token
+}
+
+
+#' Adobe S2S token response
+#'
+#' Includes the response object containing the bearer token as well as the
+#' credentials used to generate the token for seamless refreshing.
+#'
+#'
+#' @section Methods:
+#' * `refresh()`: refresh access token (if possible)
+#' * `validate()`: TRUE if the token is still valid, FALSE otherwise
+#'
+#' @docType class
+#' @keywords internal
+#' @format An R6 object
+#' @importFrom R6 R6Class
+#' @noRd
+AdobeS2SToken <- R6::R6Class("AdobeS2SToken", list(
+  secrets = NULL,
+  token = NULL,
+  initialize = function(token, secrets) {
+    self$secrets <- secrets
+    self$token <- token
+  },
+  can_refresh = function() {
+    FALSE
+  },
+  refresh = function() {
+    self$token <- auth_s2s_gen(self$secrets)
+    self
+  },
+  validate = function() {
+    self$token$expires_at + self$token$expires_at / 1000 > Sys.time() - 1200
+  }
+))
+
 
 # OAuth ----------------------------------------------------------------
 #' @family auth
